@@ -4,7 +4,7 @@
 
 /* 
  * Program to assign static device names to SCSI devices as opposed to the
- * dynamic numbering of the Linux-1.0 -- 2.4 kernel.
+ * dynamic numbering of the Linux-1.0 -- 2.6 kernel.
  * Program written by Eric Youngdale <aric@aib.com>
  * Reworked 1/2000 Kurt Garloff <garloff@suse.de>
  * 
@@ -184,13 +184,26 @@
  *       fill_in_proc()
  *   -> 2.34.
  *
- *     TODO: other char devices
- *     
  *   * 2004-05-24: Chris Townsend <Chris_Townsend@dra-hq.com>
  *     - Fix bug with channel specified in scsi.alias
  *   * 2004-06-09: Kurt Garloff <garloff@suse.de>
  *     - Handle longer partition names
  *   -> 2.35.
+ *
+ *   * 2005-05-14: Kurt Garloff <garloff@suse.de>
+ *     - For 2.4, search for more devices when finding matching sd dev.
+ *   * 2005-08-16: Kurt Garloff <garloff@suse.de>
+ *     - The 2.6 kernel now reports the 'RAID' type.
+ *     - Cleanup sysfs parsing.
+ *     - 2.6 tape support. (Thanks to David Chen <david@global.com>!)
+ *     - SG_IO support.
+ *     - Improve ID parsing (page 0x83), support EUI64 and NAA (mostly)
+ *   -> 2.36
+ *
+ *     TODO:
+ *           Change wwid to string type to handle T10 ...
+ *           handle more identifiers to match devices
+ *     
  */
 
 #include <stdio.h>
@@ -210,30 +223,35 @@
 #include <string.h>
 #include <ctype.h>
 #include <getopt.h>
-
-static char rcsid[] ="$Id: scsidev.c,v 1.28.2.55 2004/09/15 20:36:23 garloff Exp $";
-static char *versid = "scsidev " VERSION " 2004-02-09";
+ 
+static char rcsid[] ="$Id: scsidev.c,v 1.28.2.64 2005/08/17 22:55:12 garloff Exp $";
+static char *versid = "scsidev " VERSION " 2005-08-16";
 static char *copyright = "Copyright: GNU GPL  (see file COPYING)\n" \
 " (w)  1994--1997 Eric Youngdale <eric@andante.org>\n"\
-"      2000--2004 Kurt Garloff   <garloff@suse.de>";
+"      2000--2005 Kurt Garloff   <garloff@suse.de>";
 
 
 #include "config.h"
 
 #ifdef HAVE_SCSI_SCSI_H					/* { */
-#include <scsi/scsi.h>
+# include <scsi/scsi.h>
 #else							/* } { */
-#ifdef HAVE_LINUX_SCSI_H				/* { */
-#include <linux/scsi.h>
-#else							/* } { */
-#ifdef HAVE__USR_SRC_LINUX_DRIVERS_SCSI_SCSI_H 		/* { */
-#include "/usr/src/linux/drivers/scsi/scsi.h"
-#else							/* } { */
-#error "Unable to include scsi.h"
-#endif							/* } */
-#endif							/* } */
+# ifdef HAVE_LINUX_SCSI_H				/* { */
+#  include <linux/scsi.h>
+# else							/* } { */
+#  ifdef HAVE__USR_SRC_LINUX_DRIVERS_SCSI_SCSI_H 	/* { */
+#   include "/usr/src/linux/drivers/scsi/scsi.h"
+#  else							/* } { */
+#   error "Unable to include scsi.h"
+#  endif						/* } */
+# endif							/* } */
 #endif							/* } */
 
+/* SG_IO */
+#ifdef HAVE_SCSI_SG_H
+# include <scsi/sg.h>
+#endif
+
 int use_symlink = 0;
 int use_scd = 0;
 int symlink_alias = 0;
@@ -246,14 +264,15 @@ int san_del = 0;
 int no_san = 0;
 int no_procscsi = 0;
 int no_sysfs = 0;
+int full_scan = 32;
 int nm_cbtu = 0;
 int supp_rmvbl = 0;
 int supp_multi = 0;
 int override_link_perm = 1;
 char *no_serial = "No serial number";
 char *scsialias = "";
-unsigned long long no_wwid = 0;
-int no_hsv_os_id = -1;
+const unsigned long long no_wwid = 0;
+const int no_hsv_os_id = -1;
 
 #define DEVSCSI "/dev/scsi"
 #define TESTDEV DEVSCSI "/testdev"
@@ -902,7 +921,7 @@ void create_dev (sname *spnt, char symlink)
 
 /*
  * We need to "fix" any device nodes that are currently not used because
- * it is a sercurity risk to leave these lying about.  These are fixed
+ * it is a security risk to leave these laying around.  These are fixed
  * by storing a shadow file.  If these become active again,
  * we will be able to use them again because the minor number will be
  * set back again, and we are preserving the ownership and permissions.
@@ -1081,61 +1100,86 @@ int getscsiinfo (int fd, sname *spnt, int setidlun)
 
 }
 
+/* Check whether disk number no matches host/chan/id/lun in spnt */
+int comparediskidlun(sname *spnt, int no)
+{
+	int id[2];
+	int scsi_id, channel, lun, host;
+	int major, minor, res, fd;
+	
+	major = disknum_to_sd_major (no);
+	minor = (no << 4) & ~0x0f;
+    
+	res = mknod (TESTDEV, 0600 | S_IFBLK,
+		makedev (major, minor));
+	fd = open (TESTDEV, O_RDONLY | O_NONBLOCK);
+	unlink (TESTDEV);
+
+	if (fd < 0)
+		return 0;
+	res = ioctl (fd, SCSI_IOCTL_GET_IDLUN, id);
+	close (fd);
+	if (res < 0)
+		return 0;
+
+	host = id[0] >> 24;
+	channel = (id[0] >> 16) & 0xff;
+	lun = (id[0] >> 8 ) & 0xff;
+	scsi_id = id[0] & 0xff;
+
+	if (verbose >= 2) 
+		printf ("Scanning: %d==%d %d==%d %d==%d %d==%d \n",
+		host, spnt->hostnum, channel, spnt->chan,
+		scsi_id, spnt->id, lun, spnt->lun);
+	if (host == spnt->hostnum &&
+	    channel == spnt->chan &&
+	    scsi_id == spnt->id &&
+	    lun == spnt->lun) {
+		spnt->major = major;
+		spnt->minor = minor;
+		return 1;
+	}
+	return 0;
+}
+
+/* Find scsi disk high level device that points to the device
+ * defined in spnt->hostnum, chan, id, lun; search all devs
+ * lower than current disk no up to current+full_scan */
 void findscsidisk(sname * spnt, int no) 
 {
-	int status, major, minor, i, fd, res;
-	int scsi_id, channel, lun, host;
-	int id[2];
+	int i;
+	int searchln = no > full_scan? no: full_scan;
 	if (verbose >= 1) 
 		printf("Findscsidisk: %d\n",no);
-	/* only search up to 16 devices may be a bad assumption, but scanning 
-	 * the whole list could take a long time */
-	unlink (TESTDEV);
-	for (i = no; i < no+16; ++i) {
-		major = disknum_to_sd_major (i);
-		minor = (i << 4) & ~0x0f;
-     
-		status = mknod (TESTDEV, 0600 | S_IFBLK,
-				makedev (major, minor));
-		fd = open (TESTDEV, O_RDONLY);
- 		unlink (TESTDEV);
-
-		if (fd < 0)
-			continue;
-		res = ioctl (fd, SCSI_IOCTL_GET_IDLUN, id);
-		if (res < 0)
-			continue;
-
-		host = id[0] >> 24;
-		channel = (id[0] >> 16) & 0xff;
-		lun = (id[0] >> 8 ) & 0xff;
-		scsi_id = id[0] & 0xff;
-		
-		if (verbose >= 2) 
-			printf ("Scanning: %d==%d %d==%d %d==%d %d==%d \n",
-			host, spnt->hostnum, channel, spnt->chan,
-			scsi_id, spnt->id, lun, spnt->lun);
-		if (host == spnt->hostnum &&
-		    channel == spnt->chan &&
-		    scsi_id == spnt->id &&
-		    lun == spnt->lun) {
-			spnt->major = major;
-			spnt->minor = minor;
-			return;
-		}               
-	}
-	/* If we did not find a match, return a good guess.. */ 
-	//unlink (TESTDEV);       
+	/* Default values */
 	spnt->major = disknum_to_sd_major (no);
 	spnt->minor = (no << 4) & ~0x0f;
+	/* only search up to full_scan devices may be a bad assumption, but 
+	 * scanning the whole list could take a long time */
+	unlink (TESTDEV);
+
+	if (comparediskidlun(spnt, no))
+		return;
+	for (i = 1; i <= searchln; ++i) {
+		if (no-i >=0 && comparediskidlun(spnt, no-i))
+			return;
+		if (comparediskidlun(spnt, no+i))
+			return;
+	}               
+	/* If we did not find a match, return a good guess.. */ 
+	if (verbose)
+		printf("No matching disk found for %i:%i:%i:%i in 0 .. %i\n",
+			spnt->hostnum, spnt->chan, spnt->id, spnt->lun,
+			no+searchln);
 }
 
 
+/* Create dev entries for a disk */
 int build_disk (sname * spnt, int no)
 {
     int minor; int fd; int status;
     sname * spnt1 = sname_dup (spnt);
-    findscsidisk(spnt1,no);
+    findscsidisk(spnt1, no);
     spnt->partition = -1;
     spnt1->devtp = SD;
     scsiname (spnt1); oldscsiname (spnt1);
@@ -1190,7 +1234,7 @@ int build_disk (sname * spnt, int no)
 	unlink (TESTDEV);
 	if (fd < 0) 
 	    continue;
-	// TO DO: Add sanity checks here ??
+	// TODO: Add sanity checks here ??
 	close (fd);
 	spnt1 = sname_dup (spnt);
 	spnt1->partition = minor % 16;
@@ -1460,7 +1504,7 @@ void build_sgdevlist ()
     int miss = 0;	
     int minor = 0;
     int major = SCSI_GENERIC_MAJOR; 
-    int mode = O_RDWR;
+    int mode = O_RDWR | O_NONBLOCK;
     //    int devtype = (SCSI_BLK_MAJOR(major)? S_IFBLK: S_IFCHR);
     enum devtype_t devtp;
     
@@ -1530,8 +1574,10 @@ void build_sgdevlist ()
 	    case SD:
 		if (!build_disk (spnt, disks)) 
 			disks++;
+		/* This shouldn't happen, we search already in findscsidisk() 
 		else if (!build_disk (spnt, disks+1)) 
 			disks += 2;
+		 */	
 		break;
 	    case ST:
 		if (!build_tape (spnt, tapes)) 
@@ -1600,14 +1646,12 @@ int procscsi_readrecord (FILE* f)
 	return 0;
 }
 
-#ifndef TYPE_PRINTER
-# define  TYPE_PRINTER 0x02
-#endif
-#ifndef TYPE_COMM
-# define TYPE_COMM 0x09
-#endif
-
-const char *const scsi_device_types[] =
+/* This is a copy of 
+ * const char *const scsi_device_types[MAX_SCSI_DEVICE_CODE]
+ * from linux/drivers/scsi/scsi.c
+ */
+#define MAX_SCSI_DEVICE_CODE 15
+const char *const scsi_device_types[MAX_SCSI_DEVICE_CODE] =
 {
 	"Direct-Access",
 	"Sequential-Access",
@@ -1621,17 +1665,18 @@ const char *const scsi_device_types[] =
 	"Communications",
 	"Unknown",
 	"Unknown",
-	"Unknown",
+	"RAID",
 	"Enclosure",
+	"Direct-Access-RBC",
 };
 
 
 char linux_to_devtp (const char* tp)
 {
-	int idx = 0;
+	int idx;
 	if (!strcmp (tp, "Unknown"))
 		return 0x1f;
-	for (; idx < 14; ++idx)
+	for (idx = 0; idx < MAX_SCSI_DEVICE_CODE; ++idx)
 		if (!strcmp (tp, scsi_device_types[idx]))
 			return idx;
 	fprintf (stderr, "Linux kernel reports new device type \"%s\". Mail author!\n",
@@ -1937,13 +1982,14 @@ void trigger_module_loads ()
 struct sysfsdev {
 	int maj, min;
 	char nm[23];
-	char tp;
+	char blk;
 };
 
 struct sysfsdev sysfsdevs[2];
 
 
-void sysfs_get_block_nm (char* nm, struct sysfsdev* sysfsdevptr)
+/* Strip /dev part from nm and readlink the device name */
+void sysfs_get_dev_nm (char* nm, struct sysfsdev* sysfsdevptr)
 {
 	char buf[128];
 	char* ptr = strrchr (nm, '/');
@@ -1965,49 +2011,52 @@ void sysfs_get_generic_nm (char* nm, struct sysfsdev* sysfsdevptr)
 	strcpy (sysfsdevptr->nm, ptr+1);
 }
 
+int sysfs_read_devinfo(char* basenm, struct sysfsdev* sysfsptr, 
+		       sname * spnt, char* suffix, char blk)
+{
+	FILE* f;
+	char nm[128];
+	strcpy(nm, basenm);
+	strcat(nm, suffix);
+	f = fopen(nm, "r");
+	if (!f)
+		return 0;
+	sysfsptr->blk = blk;
+	fscanf(f, "%i:%i", &sysfsptr->maj, &sysfsptr->min);
+	sysfs_get_dev_nm(nm, sysfsptr);
+	if (verbose > 1)
+		printf ("sysfs_read_devinfo %d:%d:%d:%d -> %s(%c %x:%x)\n",
+			spnt->hostnum, spnt->chan, spnt->id, spnt->lun,
+			sysfsptr->nm, blk? 'b': 'c',
+			sysfsptr->maj, sysfsptr->min);
+	fclose(f);
+	return 1;
+}
+
+
 int sysfs_getinfo (sname * spnt)
 {
-	char nm[128]; char *nmptr;
-	FILE* f; int dev = 0;
+	char nm[128]; 
+	int dev = 0;
+	int inq_devtp;
+	FILE* f;
 
 	strcpy (nm, "/sys/class/scsi_device/");
 	sprintf (nm+strlen(nm), "%d:%d:%d:%d",
 		 spnt->hostnum, spnt->chan, spnt->id, spnt->lun);
 	strcat (nm, "/device/");
-	nmptr = nm+strlen(nm);
-	/* sysfs char dev */
-	*nmptr = 0;
-	strcat (nm, "block/dev");
-	f = fopen(nm, "r");
-	if (f) {
-		sysfsdevs[dev].tp = 1;
-		fscanf (f, "%i:%i", &sysfsdevs[dev].maj, &sysfsdevs[dev].min);
-		sysfs_get_block_nm (nm, sysfsdevs+dev);
-		if (verbose > 1)
-			printf ("sysfs_getinfo %d:%d:%d:%d -> %s(%x:%x)\n",
-				spnt->hostnum, spnt->chan, spnt->id, spnt->lun,
-				sysfsdevs[dev].nm, sysfsdevs[dev].maj, sysfsdevs[dev].min);
-		++dev;
-		fclose (f);
-	}
-	// Bad luck: linux-2.6-test1 does not display sg devices
-	//strcat (nm, "char/dev");
-	*nmptr = 0;
-	strcat (nm, "generic");
-	f = fopen(nm, "r");
-	if (f) {
-		sysfsdevs[dev].tp = 0;
-		sysfs_get_generic_nm (nm, sysfsdevs+dev);
-		//fscanf (f, "%i:%i", &sysfsdevs[dev].maj, &sysfsdevs[dev].min);
-		sysfsdevs[dev].maj = SCSI_GENERIC_MAJOR;
-		sscanf (sysfsdevs[dev].nm, "sg%i", &sysfsdevs[dev].min);
-		if (verbose > 1)
-			printf ("sysfs_getinfo %d:%d:%d:%d -> %s(%x:%x)\n",
-				spnt->hostnum, spnt->chan, spnt->id, spnt->lun,
-				sysfsdevs[dev].nm, sysfsdevs[dev].maj, sysfsdevs[dev].min);
-		++dev;
-		fclose (f);
-	}
+	/* sysfs devs */
+	dev += sysfs_read_devinfo(nm, sysfsdevs+dev, spnt, "block/dev", 1);
+	dev += sysfs_read_devinfo(nm, sysfsdevs+dev, spnt, "tape/dev", 0);
+	dev += sysfs_read_devinfo(nm, sysfsdevs+dev, spnt, "generic/dev", 0);
+	
+	strcat (nm, "type");
+	f = fopen (nm, "r");
+	fscanf (f, "%i", &inq_devtp);
+	spnt->inq_devtp = inq_devtp;
+	/* TODO: Use much more info from sysfs, e.g. driver */
+	fclose (f);
+	
 	return dev;
 }	
 
@@ -2020,9 +2069,13 @@ void sysfs_parse (sname *spnt, int idx)
 		*p2++ = *p1++;
 	*p2 = 0;
 	spnt->oldname = strdup(sysfsdevs[idx].nm);
-	spnt->major = sysfsdevs[idx].maj;
-	spnt->minor = sysfsdevs[idx].min;
-	spnt->devtp = nm26_to_devtp(nm);
+	spnt->major   = sysfsdevs[idx].maj;
+	spnt->minor   = sysfsdevs[idx].min;
+	//spnt->blk     = sysfsdevs[idx].blk;
+	/* Note: We need to handle generic here, thus can't rely on inq_devtp */
+	//spnt->devtp   = inq_devtp_to_devtp(spnt->inq_devtp, spnt);
+	spnt->devtp   = nm26_to_devtp(nm);
+
 	if (verbose >= 1) {
 		dumpentry(spnt);
 		printf("Names %s, %s\n", spnt->oldname, nm);
@@ -2252,9 +2305,9 @@ int main (int argc, char * argv[])
     }
     while ((c = getopt(argc, argv, "ypflLvqshnderoMm:c:A:")) != -1) {
 	switch (c) {
-	  case 'y':
+	  case 'y':	/* undocumented */
 	    no_sysfs = 1; break;
-	  case 'p':
+	  case 'p':	/* undocumented */
 	    no_procscsi = 1; break;
 	  case 'f':
 	    force = 1; break;
@@ -2838,7 +2891,7 @@ void dumppage (unsigned char* page)
 {
     int ln = 4 + page[3];
     int i;
-    for (i = 0; i <= ln; i++)
+    for (i = 0; i < ln; i++)
     {
 	printf (" %02x", page[i]);
 	if (!((i+1)%16)) printf ("\n");
@@ -2849,9 +2902,9 @@ void dumppage (unsigned char* page)
 
 void dumppage_83 (unsigned char* page)
 {
-    int ln = (page[2] << 8) + page[3];
+    int ln = 4 + (page[2] << 8) + page[3];
     int i;
-    for (i = 0; i <= ln; i++)
+    for (i = 0; i < ln; i++)
     {
 	printf (" %02x", page[i]);
 	if (!((i+1)%16)) printf ("\n");
@@ -2881,74 +2934,161 @@ char* getstr (char* page, int start, int stop)
 	return str;
 }
 
+struct pg83id {
+	unsigned char pident;
+	unsigned char idtype;
+	unsigned char res;
+	unsigned char idlen;
+	unsigned char iddata[8];
+};
+
+struct pg83 {
+	unsigned char periph;
+	unsigned char pgcode;
+	unsigned short length;	/* big endian ! */
+	struct pg83id idlist;
+};
+	
+unsigned long long get_no(unsigned char* data, 
+		          unsigned char start, unsigned char stop)
+{
+	unsigned int dat = ntohl(*(unsigned int*) data);
+	/* Clear left bits if number does not start at 0 */
+	dat  &= (0xffffffff >> start);
+	/* Shift right if stop is not 32bits away */
+	dat >>= (32-stop);
+	return dat;
+}
+
+
+/* Return the wwid as 64bit number, high 32bits have vendor ID (IEEE)
+ * and low have device ID */
 unsigned long long extract_wwid (unsigned char* page)
 {
-	unsigned int hi, lo;
-	char vendor_id[9];
-	char *device_id;
-
-	if (page[1] != 0x83) {
+	struct pg83 *p83 = (struct pg83*) page;
+	struct pg83id *pid = &p83->idlist;
+	unsigned int len = ntohs(p83->length);
+	
+	if (p83->pgcode != 0x83) {
 	    printf("Page does not contain any WWID\n");
 	    return no_wwid;
 	}
-	/* We only support binary data */
-	if ((page[4] & 0x0f) == 1) {
-	    /* We support identifier types 2, 3 */
-	    if ((page[5] & 0x0f) != 3 && (page[5] & 0x0f) != 2)
-		return no_wwid;
-	    /* Id length should be 8 (64 bits) */
-	    if (page[7] != 8)
-		return no_wwid;
-            /* ASSOCIATION must be with the device (value 0) */ 
-	    if ((page[5] & 0x30) != 0)
-		return no_wwid;
-	    /* Now translate network byte order to host byte order */
-	    hi = ntohl (*(unsigned int*)(page+ 8));
-	    lo = ntohl (*(unsigned int*)(page+12));
-	    return ((unsigned long long) hi << 32) + (unsigned long long) lo;
-	} else if ((page[4] & 0x0f) == 2) {
-	    /* Get ASCII data */
-	    if ((page[5] & 0x0f) == 1) {
-		/* T-10 Identifier */
-		memcpy(vendor_id, page+8, 8);
-		vendor_id[9] = '\0';
-		if (verbose >= 1)
-		    printf("Vendor: %s\n",vendor_id);
-		if (page[7] > 8) {
-		    device_id = malloc(page[7]-7);
-		    device_id[page[7]-8] = '\0';
-		    memcpy(device_id, page+16, page[7] - 8);
-		    if (verbose >= 1)
-			printf("Device: %s\n", device_id);
-		    free(device_id);
-		}
-	    }
-	    return no_wwid;
-	} else {
-	    return no_wwid;
+
+	if (pid->res != 0) {
+		/* Pre SCSI-2 / pre-SPC answer */
+		return get_no(&pid->pident  , 0, 32) << 32
+		     | get_no(&pid->pident+4, 0, 32);
 	}
+	
+	for (;
+	     (char*)pid < (char*)p83 + 4 + len;
+	     pid = (struct pg83id*) ((char*)pid + 4 + pid->idlen)) {
+		/* SKip over per target of per port association */
+		unsigned char assoc  = (pid->idtype & 0x30) >> 4;
+		unsigned char code   =  pid->pident & 0x0f;
+		unsigned char idtype =  pid->idtype & 0x0f;
+		if (assoc != 0)
+			continue;
+		/* EUI-64, should always be binary, */
+		if (idtype == 2 && code  == 1) {
+			if (pid->idlen == 8 || pid->idlen == 12 
+					    || idtype == 3)
+				return get_no(pid->iddata  , 0, 32) << 32 
+				     | get_no(pid->iddata+4, 0, 32);
+			else if (pid->idlen == 16)
+				return get_no(pid->iddata+4, 0, 32) << 32
+				     | get_no(pid->iddata+8, 0, 32);
+			else continue;
+		}
+		/* NAA */
+		if (idtype == 3 && code == 1) {
+			unsigned char naa = (pid->iddata[0] & 0xf0) >> 4;
+			if (naa == 2) 
+				return  get_no(pid->iddata+2, 0, 24) << 32
+				     | (get_no(pid->iddata  , 4, 12) & 0xff) << 24
+				     |  get_no(pid->iddata+5, 0, 24);
+			else if (naa == 5 || naa == 6)
+				return get_no(pid->iddata  , 4, 28) << 36
+				     | get_no(pid->iddata+3, 4,  8) << 32
+				     | get_no(pid->iddata+4, 0, 32);
+			else 
+				continue;
+		}
+		/* We can't really handle T10, it's ASCII/UTF8 */
+		if (idtype == 1 && code == 2 && verbose) {
+			char  ven_id[10], dev_id[130];
+			memcpy(ven_id, pid->iddata, 8);
+			ven_id[9] = 0;
+			memcpy(dev_id, pid->iddata+8, pid->idlen-8);
+			dev_id[pid->idlen] = 0;
+			printf("T10 ID: \"%s\" \"%s\"\n", ven_id, dev_id);
+			continue;
+		}
+	}
+	return no_wwid;
+}
+
+#ifndef SG_IO
+void my_memmove(unsigned char* dst, unsigned char* src, unsigned int ln)
+{
+	if (src > dst) {
+		while (ln--)
+			*dst++ = *src++;
+	} else {
+		dst += ln; src += ln;
+		while (ln--)
+			*(--dst) = *(--src);
+	}
+}
+#endif
+
+int scsi_cmd(int file, int rlen,
+	     unsigned char* cmd, int cmdlen, 
+	     unsigned char* buf, int buflen,
+	     unsigned char* sen, int senlen)
+{
+	int ret;
+#ifdef SG_IO
+	sg_io_hdr_t sghdr;
+	memset(&sghdr, 0, sizeof(sg_io_hdr_t));
+	sghdr.interface_id = 'S';
+	sghdr.dxfer_direction = SG_DXFER_FROM_DEV;
+	sghdr.cmd_len = cmdlen;
+	sghdr.iovec_count = 0;
+	sghdr.dxfer_len = rlen;
+	sghdr.dxferp = buf;
+	sghdr.cmdp = cmd;
+	sghdr.mx_sb_len = senlen;
+	sghdr.sbp = sen;
+	sghdr.timeout = 2000;
+	if (sen)
+		memset(sen, 0, senlen);
+	memset(buf, 0, buflen);
+
+	ret = ioctl(file, SG_IO, &sghdr);
+	if (verbose >= 2)
+		printf("SG_IO %02x %02x %02x: ret=%i, status=%i (host %i, drv %i), read=%i/%i\n",
+		       cmd[0], cmd[1], cmd[2],	
+		       ret, sghdr.status, sghdr.host_status, sghdr.driver_status,
+		       rlen-sghdr.resid, rlen);
+	return ret + sghdr.status;
+#else
+	memset(buf, 0, buflen);
+	*(  (int *) buf)     = 0;	/* Length of input data */
+	*( ((int *) buf+1) ) = rlen;	/* Length of output data */
+	memcpy(buf+8, cmd, cmdlen);
+	ret = ioctl(file, SCSI_IOCTL_SEND_COMMAND, buf);
+	my_memmove(buf, buf+8, buflen-8);
+	return ret;
+#endif
 }
 
 #define INQBUFSZ 512
 int get_inq_page (int file, int lun, unsigned char* buf, unsigned char page, char evpd)
 {
-	unsigned char *cmd;
-	
-	memset (buf, 0, INQBUFSZ);
-	
-	*( (int *)  buf)	= 0;			/* length of input data */
-	*( ((int *) buf) + 1 )	= INQBUFSZ-32;	/* length of output buffer */
-	
-	cmd = (char *) ( ((int *) buf) + 2 );
-	
-	cmd[0] = 0x12;			/* INQUIRY */
-	cmd[1] = (lun << 5) | (evpd? 1: 0);	/* lun, evpd=... */
-	cmd[2] = page;			/* page code 0: std inquiry */
-	cmd[3] = 0x00;			/* (reserved) */
-	cmd[4] = 0xfc;			/* allocation length */
-	cmd[5] = 0x00;			/* control */
-	
-	return ioctl (file, SCSI_IOCTL_SEND_COMMAND, buf);
+	unsigned char cmd[6] = { 0x12, (lun << 5 ) | (evpd? 1: 0), page,
+				 0x00, 0xfc, 0x00 };
+	return scsi_cmd(file, 0xfc, cmd, 6, buf, INQBUFSZ, NULL, 0);
 }
 	
 int inquiry (int infile, sname * spnt)
@@ -2970,7 +3110,7 @@ int inquiry (int infile, sname * spnt)
 #else
     int status;
     unsigned char buffer[INQBUFSZ];
-    unsigned char * const pagestart = buffer + 8;
+    unsigned char * const pagestart = buffer;// + 8;
     //int infile;
     char have_ser_page = 0;
     char have_wwid_page = 0;
@@ -2992,10 +3132,8 @@ int inquiry (int infile, sname * spnt)
 		 spnt->name, spnt->id, spnt->lun, spnt->major, spnt->minor);
 	return -1;
     }
-    if (verbose >= 2) {
-	printf("Calling dumppage\n");
-	dumppage (pagestart);
-    }
+    /* dumppage does not make sense for std INQUIRY */
+
     spnt->manufacturer = getstr (pagestart, 8, 15); 
     spnt->model = getstr (pagestart, 16, 31);
     spnt->rev = getstr (pagestart, 32, 35);
@@ -3060,40 +3198,24 @@ int inquiry (int infile, sname * spnt)
 int get_hsv_os_id (int infile, sname * spnt)
 {
   int status,i;
-  unsigned char *cmd;
-  unsigned char *pagestart;
   unsigned char buffer[1024];
+  unsigned char *pagestart = buffer;
   char path[64];
   
+  unsigned char cmd[12] = {
+	  0xa3, 0x05, 0x00 /*res*/, 0x00 /*res*/,
+	  0x00 /* lun ign */, 0x00 /* lun ign */,
+	  0x00 /* len */, 0x00 /* len */, 0x00 /* len */, 0xfc /* len */,
+	  0x00, 0x00 };
+  
   spnt->hsv_os_id = no_hsv_os_id;
 
   if (!spnt->model || strncmp (spnt->model, "HSV", 3) != 0)
        return -1;
-  memset (buffer, 0, sizeof(buffer));
-  
-  *( (int *)  buffer )         = 0;    /* length of input data */
-  *( ((int *) buffer) + 1 )    = sizeof(buffer)-16;    /* length of output buffer */
-
-  cmd = (char *) ( ((int *) buffer) + 2 );
-  
-  cmd[0] = 0xa3;
-  cmd[1] = 0x05;
-  cmd[2] = 0x00;	//res          
-  cmd[3] = 0x00;	//res  
-  cmd[4] = 0x00;	//lun ignored
-  cmd[5] = 0x00;	//lun ignored
-  cmd[6] = 0x00;	//len
-  cmd[7] = 0x00;	//len
-  cmd[8] = 0x00;	//len
-  cmd[9] = 0xff;	//len
-  cmd[10] = 0x00;	//PRUCLU
-  cmd[11] = 0x00;
-  
   if( infile == -1 ) 
 	return -1;
-
-  status = ioctl( infile,  SCSI_IOCTL_SEND_COMMAND , buffer );
-  pagestart = buffer + 8;
+  
+  status = scsi_cmd(infile, 0xfc, cmd, 12, buffer, 1024, NULL, 0);
 
   if (status) 
 	spnt->hsv_os_id = no_hsv_os_id;
@@ -3104,8 +3226,8 @@ int get_hsv_os_id (int infile, sname * spnt)
 	printf ("HSV OS Unit ID for %s: %d\n", path, spnt->hsv_os_id);
 
   if (verbose == 2) {
-	for (i = 0; i <= 16; i++) {
-		printf (" %02x", buffer[i]);
+	for (i = 0; i < 16; i++) {
+		printf (" %02x", pagestart[i]);
 		if (!((i+1)%16)) 
 			printf ("\n");
 	}
